#include "tmesh.h"
#include "v3.h"

#include <fstream>
#include <iostream>

#include <float.h>

TMesh::TMesh() : vertices(0), verticesN(0), colors(0), connectivity(0),
  trianglesN(0) {}

void TMesh::Load(char *fname) {

  ifstream ifs(fname, ios::binary);
  if (ifs.fail()) {
    cerr << "INFO: cannot open file: " << fname << endl;
    return;
  }

  ifs.read((char*)&verticesN, sizeof(int));
  char yn;
  ifs.read(&yn, 1); // always xyz
  if (yn != 'y') {
    cerr << "INTERNAL ERROR: there should always be vertex xyz data" << endl;
    return;
  }
  if (vertices)
    delete vertices;
  vertices = new V3[verticesN];
  projVertices = new V3[verticesN];

  ifs.read(&yn, 1); // colors 3 floats
  if (colors)
    delete colors;
  colors = 0;
  if (yn == 'y') {
    colors = new V3[verticesN];
  }

  ifs.read(&yn, 1); // normals 3 floats
  V3 *normals = 0; // don't have normals for now
  if (normals)
    delete normals;
  normals = 0;
  if (yn == 'y') {
    normals = new V3[verticesN];
  }

  ifs.read(&yn, 1); // texture coordinates 2 floats
  float *tcs = 0; // don't have texture coordinates for now
  if (tcs)
    delete tcs;
  tcs = 0;
  if (yn == 'y') {
    tcs = new float[verticesN*2];
  }

  ifs.read((char*)vertices, verticesN*3*sizeof(float)); // load vertices

  if (colors) {
    ifs.read((char*)colors, verticesN*3*sizeof(float)); // load colors
  }

  if (normals)
    ifs.read((char*)normals, verticesN*3*sizeof(float)); // load normals

  if (tcs)
    ifs.read((char*)tcs, verticesN*2*sizeof(float)); // load texture coordinates

  ifs.read((char*)&trianglesN, sizeof(int));
  if (connectivity)
    delete connectivity;
  connectivity = new unsigned int[trianglesN*3];
  ifs.read((char*)connectivity, trianglesN*3*sizeof(unsigned int)); // read tiangles

  ifs.close();

  cerr << "INFO: loaded " << verticesN << " verts, " << trianglesN << " tris from " << endl << "      " << fname << endl;
  cerr << "      xyz " << ((colors) ? "rgb " : "") << ((normals) ? "nxnynz " : "") << ((tcs) ? "tcstct " : "") << endl;

  if (normals)
    delete normals;
  if (tcs)
    delete tcs;

}

void TMesh::RenderPoints(PPC *ppc, FrameBuffer *fb) {

  for (int i = 0; i < verticesN; i++) {
    if (projVertices[i][0] == FLT_MAX)
      continue;
    fb->SetIfCloser(projVertices[i], colors[i]);
  }

}

void TMesh::Project(PPC *ppc) {

  for (int i = 0; i < verticesN; i++) {
    ppc->Project(vertices[i], projVertices[i]);
  }

}

void TMesh::Render(PPC *ppc, FrameBuffer *fb) {

  Project(ppc);

  for (int tri = 0; tri < trianglesN; tri++) {
    unsigned int inds[3] = {connectivity[3*tri+0], 
      connectivity[3*tri+1], connectivity[3*tri+2]};
    V3 pvs[3] = {projVertices[inds[0]], projVertices[inds[1]], projVertices[inds[2]]};

    // return if any vertex behind head
    if (pvs[0][0] == FLT_MAX)
      continue;
    if (pvs[1][0] == FLT_MAX)
      continue;
    if (pvs[2][0] == FLT_MAX)
      continue;

    // compute bounding box
    BBox bbox;
    bbox.AddPoint(pvs[0]);
    bbox.AddPoint(pvs[1]);
    bbox.AddPoint(pvs[2]);
    BBox imageFrame;
    float border = 0.0f;
    imageFrame.AddPoint(V3(0.0f+border, 0.0f+border, -FLT_MAX)); 
    imageFrame.AddPoint(V3((float)fb->w-1.0f-border, (float)fb->h-1.0f-border, FLT_MAX));
    if (!bbox.Clip(imageFrame))
      continue;
    int left = (int) (bbox.corners[0][0]+0.5f);
    int right = (int) (bbox.corners[1][0]-0.5f);
    int top = (int) (bbox.corners[0][1]+0.5f);
    int bottom = (int) (bbox.corners[1][1]-0.5f);

    // rasterization setup
    M3x3 barycentricInverse = ComputeBarycentricInverse(pvs);
    M3x3 ssiM = ComputeSSI(pvs);
    V3 zABC = ssiM*V3(pvs[0][2], pvs[1][2], pvs[2][2]);
    M3x3 cols;
    cols[0] = colors[inds[0]];
    cols[1] = colors[inds[1]];
    cols[2] = colors[inds[2]];
    M3x3 colABC = (ssiM*cols);
    colABC.Transpose();

    for (int v = top; v <= bottom; v++) {
      for (int u = left; u <= right; u++) {
        V3 pixv(.5f+(float)u, .5f+(float)v, 1.0f);
        V3 baryCoords = barycentricInverse*pixv;
        if (baryCoords[0] < 0.0f || baryCoords[1] < 0.0f ||
          baryCoords[2] < 0.0f)
          continue;
        float currz = zABC*pixv;
        V3 currCol = colABC*pixv;
        fb->SetIfCloser(V3(pixv[0], pixv[1], currz), currCol);
      }
    }
  }

}

V3 TMesh::GetCenter() {

  V3 ret(0.0f, 0.0f, 0.0f);
  for (int i = 0; i < verticesN; i++) {
    ret = ret + vertices[i];
  }
  ret = ret / (float) verticesN;
  return ret;

}

void TMesh::Translate(V3 trans) {

  for (int i = 0; i < verticesN; i++) {
    vertices[i] = vertices[i] + trans;
  }

}

void TMesh::Scale(float scf) {

  for (int i = 0; i < verticesN; i++) {
    vertices[i] = vertices[i]*scf;
  }

}

BBox TMesh::ComputeBBox() {

  BBox ret;
  for (int i = 0; i < verticesN; i++)
    ret.AddPoint(vertices[i]);

  return ret;

}

void TMesh::PlaceAndScale(V3 newCenter, 
                          float bboxDiag) {

  V3 center = GetCenter();
  BBox bbox = ComputeBBox();
  Translate(center*-1.0f);
  Scale(bboxDiag/
    (bbox.corners[0]-bbox.corners[1]).Length());
  Translate(newCenter);

}

M3x3 TMesh::ComputeBarycentricInverse(V3 pvs[3]) {

  M3x3 ret;
  M3x3 vM;
  vM[0] = V3(pvs[0][0], pvs[1][0], pvs[2][0]);
  vM[1] = V3(pvs[0][1], pvs[1][1], pvs[2][1]);
  vM[2] = V3(1.0f, 1.0f, 1.0f);
  ret = vM.Invert();
  return ret;

}

M3x3 TMesh::ComputeSSI(V3 pvs[3]) {

  M3x3 ret;
  M3x3 vM;
  vM[0] = pvs[0]; vM[0][2] = 1.0f;
  vM[1] = pvs[1]; vM[1][2] = 1.0f;
  vM[2] = pvs[2]; vM[2][2] = 1.0f;
  ret = vM.Invert();
  return ret;

}